/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: utdate.h 23 2003-10-28 02:24:16Z drory $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *                         Claus A. Goessl <cag@usm.uni-muenchen.de>
 *                         Arno Riffeser <arri@usm.uni-muenchen.de>
 *                         Jan Snigula <snigula@usm.uni-muenchen.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */


// class for many time specific functions
// all output is UTC
// don't get confused!
// time is kept in "time_t ut_date" and always is UTC
// all constructors are built to get UTC

// valid for times between Jan. 1st 1970 and
// Dec. 31st 2037 due to overrun of epoche (Jan. 2038)
// Exception type DateException are thrown on error

// Date class
// DCF-string conversion only for input

#ifndef __UTDATE_H__
#define __UTDATE_H__

#include <ltl/config.h>

#include <string>
#include <iostream>
#include <iomanip>

#ifdef HAVE_SSTREAM
#ifdef HAVE_IOS

#define UTIL_USING_SSTREAM
#include <sstream>
#ifdef UTIL_USING_NAMESPACE
using std::stringstream;
using std::ios_base;
using std::ends;
#endif

#endif
#endif

#ifndef UTIL_USING_SSTREAM

#ifdef HAVE_STRSTREAM
#include <strstream>
#ifdef UTIL_USING_NAMESPACE
using std::strstream;
using std::ios;
using std::ends;
#endif

#else
#error <sstream> or <strstream> needed!
#endif

#endif

#ifdef TM_IN_SYS_TIME
#  include <sys/time.h>
#else
#  include <ctime>
#endif

#include <ltl/util/u_exception.h>

#ifdef UTIL_USING_NAMESPACE
using std::string;
using std::setprecision;
#endif

UTIL_BEGIN_NAMESPACE

// mother:
class UTDate;
// input methods:
// constructor (void): get actual UT
// constructor (time_t): assume time_t to hold UT
// output methods:
// toString(): return UTC date string

// daughters:

class DCFDate;
// input methods:
// constructor (void): get actual UT
// constructor (string): interpret DCF-77 time string
// constructor (class UTDate): initialise from UTDate or heirs
// output methods:
// toString(): return UTC date string

class FitsDate;
// input methods:
// constructor (void): get actual UT
// constructor (string): interpret FITS DATE string (already UTC)
// constructor (class UTDate): initialise from UTDate or heirs
// output methods:
// toString(): return UTC string in FITS DATE format

class JulDate;
// input methods:
// constructor (void): get actual UT
// constructor (string): interpret string as Julian Date
// constructor (double): interpret double as Julian Date
// constructor (class UTDate): initialise from UTDate or heirs
// output methods:
// toString(): return Julian date string
// toDouble(): return Julian date double

class UTTime;

class GSTTime;

class UTDate
{
   protected:
      char* toCString()const;
      time_t ut_date_;
      time_t mkuttime(struct tm* ut_tm);
      time_t mkdcftime(struct tm* dcf_tm);
   public:
      UTDate() : ut_date_(time(NULL))
      { };
      UTDate( const time_t& init_date ) : ut_date_(init_date)
      { };

      string toString()const;
};

class DCFDate : public UTDate
{
   public:
      DCFDate() : UTDate()
      { };
      DCFDate( const UTDate& d ) : UTDate(d)
      { };
      DCFDate(const string& init_date);
      string toString()const;
};

class FitsDate : public UTDate
{
      bool shortformat_;
   public:
      FitsDate() : UTDate(), shortformat_(false)
      { };
      FitsDate( const UTDate& d) : UTDate(d), shortformat_(false)
      { };
      FitsDate( const string& init_date);

      string toString()const;
};

class JulDate : public UTDate
{
   protected:
      static const double juldate_on_epoch_;    //=2440587.5;
      static const double seconds_of_day_;      //=86400.0;
      static const double juldate_end_epoch_;   //=2465424.5;
      time_t toTime_t(const double& init_date)const;
      time_t toTime_t(const string& init_date)const;
   public:
      JulDate() : UTDate()
      { };
      JulDate( const UTDate& d) : UTDate(d)
      { };
      JulDate( const double init_date)
      {
         ut_date_ = toTime_t(init_date);
      };
      JulDate( const string& init_date)
      {
         ut_date_ = toTime_t(init_date);
      };

#ifdef HAVE_SSTREAM
#ifdef HAVE_IOS
      string toString( const int prec=15 )const;
#else
      string toString( const int prec=8 )const;
#endif
#endif
      double toDouble()const;
};

class UTTime : public UTDate
{
   public:
      UTTime() : UTDate()
      { };
      UTTime( const UTDate& d ) : UTDate(d)
      { };
      UTTime( const string& t );

      string toString()const;
};

class GSTime : public UTDate
{
   public:
      GSTime() : UTDate()
      { };
      GSTime( const UTDate& d ) : UTDate(d)
      { };
     
      string toString()const;
};

UTIL_END_NAMESPACE

#ifdef UTIL_USING_SSTREAM
#undef UTIL_USING_SSTREAM
#endif

#endif
