/* 
 * apphot.cpp
 * elliptical aperture photometry
 *
 * --------------------------------------------------------------------
 * $Id: apphot.cpp 37 2004-10-06 21:12:28Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <list>


#include <iostream>
#include <iomanip>
using std::setw;

#include <cmath>
#include <ltl/marray.h>
#include <ltl/marray_io.h>
#include <ltl/statistics.h>


#include "yoda.h"
#include "photdata.h"
#include "mathutil.h"
#include "globals.h"

typedef MArray<float,2> MImage;
typedef MArray<char, 2> MMask;
typedef MArray<float,1> MVector;

using namespace ltl;

extern Globals g;


bool check_variance( const MImage& I );

float integrate_soft_e_mask( const MImage& I, const float sky, 
                             float xc, float yc,
                             float a, float b, float pa,
                             const int subsample, float *area );

float integrate_soft_e_mask_fixed( const MImage& I, const float sky, 
                                   float xc, float yc,
                                   float ap, 
                                   const int subsample, float *area );

MImage sky_e_mask( const MImage& I, const float xc, const float yc,
                   const float a1, const float b1,  
                   const float a2, const float b2,  
                   const float pa, const float nanval );



void kron_aperture_phot( const MImage& Image, list<PhotData*>& pl,
                         const float k,  const float minap, 
                         const float sky1, const float sky2,
                         const float gain, const float magzero, 
                         const float nanval )
{
   list<PhotData*>::iterator i = pl.begin();
   int nborder = 0, nbadregion = 0, count=0;
   
   // should we use a local sky estimate or a value based on
   // interpolating the global background estimate ?
   const bool local_sky = (sky1 > 0.5) && (sky2 > 0.5);

   for( ; i!=pl.end(); ++i ) 
   {
      if( !(++count%100) )
         cerr << setw(7) << count << " "; 

      PhotData& pd = **i;
      
      const int xc = int(pd.xc_ + 0.5);
      const int yc = int(pd.yc_ + 0.5);

      // we are doing a kron-like elliptical aperture and a fixed circular
      // aperture

      // check against minimal aperture size
      const float se = sqrt( pd.a_/pd.b_ );
      if( k*se*pd.ir1_ < minap )
         pd.ir1_ = minap/k/se;
      if( k/se*pd.ir1_ < minap )
         pd.ir1_ = minap/k*se;

      pd.a_ = k*se*pd.ir1_;
      pd.b_ = k/se*pd.ir1_;
      
      // find enclosing rectangle
      const float ax = fabs( pd.ir1_*se * cos(pd.pa_rad()) );
      const float bx = fabs( pd.ir1_/se * sin(pd.pa_rad()) );
      const float ay = fabs( pd.ir1_*se * sin(pd.pa_rad()) );
      const float by = fabs( pd.ir1_/se * cos(pd.pa_rad()) );

      int wx = ax >= bx ? int(ax*k + 0.5) : int(bx*k + 0.5);
      int wy = ay >= by ? int(ay*k + 0.5) : int(by*k + 0.5);
      
      // add space for sky annulus
      wx += int(sky1 + sky2 + 1);
      wy += int(sky1 + sky2 + 1);
      
      // the reference to the subimage
#ifdef YODA_DEBUG_PHOTOMETRY
      cerr << "Photometry of : ";
      cerr << xc << "+-" << wx << "  ";
      cerr << yc << "+-" << wy << "   ";
      cerr << "A:" << pd.a_ << " B:" << pd.b_ << " PA:"<< pd.pa_ << endl;
#endif
      // check that the enclosing rectangle is within the image
      if( xc-wx<Image.minIndex(1) || xc+wx>Image.maxIndex(1) ||
          yc-wy<Image.minIndex(2) || yc+wy>Image.maxIndex(2) )
      {
         pd.f_ = pd.df_ = pd.m_ = pd.dm_ = pd.sky_ = pd.dsky_ = -99.;
         pd.ff_ = pd.dff_ = pd.fm_ = pd.dfm_ = -99.;
#ifdef YODA_DEBUG_PHOTOMETRY
         cerr << "Object at border .. " << xc << " " << yc << endl;
#endif
         ++nborder;
         continue;
      }
         
      MImage I = Image( Range(xc-wx, xc+wx), Range(yc-wy, yc+wy) );
        
      // check to see if data look reaonable ...
      if( !check_variance( I ) )
      {
         pd.f_ = pd.df_ = pd.m_ = pd.dm_ = pd.sky_ = pd.dsky_ = -99.;
         pd.ff_ = pd.dff_ = pd.fm_ = pd.dfm_ = -99.;
         ++nbadregion;
#ifdef YODA_DEBUG_PHOTOMETRY
         cerr << "Object in bad region .. " << xc << " " << yc << endl;
#endif
         continue;
      }
         
      // if everything is ok, we can perform the photometry ...

      // (subarray has coordinates 1..2wx+1 => new center of ap
      // in subimage is the fractional part of the coordinate + wx+1      
      float xcs = (pd.xc_-xc)+(wx+1);
      float ycs = (pd.yc_-yc)+(wy+1);

      // calculate sky
      double sky, sigma_sky;
      int nsky; 

      if( local_sky )
      {
         // the sky region is simply a difference of two elliptical masks ...
         MImage Sky = sky_e_mask( I, xcs, ycs, pd.a_+sky1+sky2, pd.b_+sky1+sky2, 
                               pd.a_+sky1, pd.b_+sky1, pd.pa_, nanval );
         // estimate a local sky value from an annulus around the object
         nsky = kappa_sigma_average( Sky, 2.5, 0.0f, &sky, &sigma_sky );
      }
      else
      {
         nsky = int(g.bggrid_*g.bggrid_*0.5);
         sky = bilin_interp( g.bg_, int(xc+0.5), int(yc+0.5), g.bggrid_ );
         sigma_sky=bilin_interp( g.rms_, int(xc+0.5), int(yc+0.5), g.bggrid_ );
      }

      // for the object, a smaller subimage is sufficient ...
      wx -= int(sky1+sky2);
      wy -= int(sky1+sky2);
      MImage A = Image( Range(xc-wx, xc+wx), Range(yc-wy, yc+wy) );
      xcs = (pd.xc_-xc)+(wx+1);
      ycs = (pd.yc_-yc)+(wy+1);
      
      // integrate flux-sky within subsampled aperture
      // for kron-like aperture
      float area=0.0;
      float flux=0.0;
      float skyerr2=0.0;
      float objerr2=0.0;
      float ferr=0.0;

      flux = integrate_soft_e_mask( A, sky, 
                                    xcs, ycs,
                                    pd.a_, pd.b_, pd.pa_, 
                                    5, 
                                    &area );
      // error in sky estimate
      skyerr2 = area*area*sigma_sky*sigma_sky/nsky;
      // poisson error in flux + background noise
      objerr2 = (flux>0.0 ? flux/gain : 0.0) + 
         area*sigma_sky*sigma_sky;
      // total error
      ferr = sqrt( skyerr2 + objerr2 );
         
#ifdef YODA_DEBUG_PHOTOMETRY
      cerr << "   F:" << flux << " dF:" << ferr 
           << " S:" << sky << " dS:" << sqrt(skyerr2) << endl;
#endif
         
      pd.f_ = flux;
      pd.df_ = ferr;
      pd.sky_ = sky;
      pd.dsky_ = sigma_sky;
         
      if( flux > 0.0 )
      {
         pd.m_  = -2.5*log10( pd.f_ ) + magzero;
         pd.dm_ =  2.5/M_LN10/pd.f_*pd.df_;
      }
      else
      {
         pd.m_  = 99.;
         pd.dm_ = 99;
      }
      
   }
   cerr << endl << "  " << nborder << " objects too close to or outside image border.\n";
   cerr << "  " << nbadregion << " objects in bad image regions.\n";
}



void fixed_aperture_phot( const MImage& Image, list<PhotData*>& pl,
                          const float aper, 
                          const float sky1, const float sky2,
                          const float gain, const float magzero, 
                          const float nanval )
{
   list<PhotData*>::iterator i = pl.begin();
   int nborder = 0, nbadregion = 0, count=0;

   // should we use a local sky estimate or a value based on
   // interpolating the global background estimate ?
   const bool local_sky = (sky1 > 0.5) && (sky2 > 0.5);

   for( ; i!=pl.end(); ++i ) 
   {
      if( !(++count%100) )
         cerr << setw(7) << count << " "; 

      PhotData& pd = **i;
      
      const int xc = int(pd.xc_ + 0.5);
      const int yc = int(pd.yc_ + 0.5);

      // we are doing a kron-like elliptical aperture and a fixed circular
      // aperture

      // find enclosing rectangle
      int wx = int(aper+0.5);
      int wy = int(aper+0.5);
      
      // add space for sky annulus
      wx += int(sky1 + sky2 + 1);
      wy += int(sky1 + sky2 + 1);
      
      pd.fixed_ = aper;

      // the reference to the subimage
#ifdef YODA_DEBUG_PHOTOMETRY
      cerr << "Photometry of : ";
      cerr << xc << "+-" << wx << "  ";
      cerr << yc << "+-" << wy << "   ";
      cerr << "R:" << pd.fixed_ << endl;
#endif
      // check that the enclosing rectangle is within the image
      if( xc-wx<Image.minIndex(1) || xc+wx>Image.maxIndex(1) ||
          yc-wy<Image.minIndex(2) || yc+wy>Image.maxIndex(2) )
      {
         pd.ff_ = pd.dff_ = pd.fm_ = pd.dfm_ = pd.fsky_ = pd.dfsky_ = -99.;
#ifdef YODA_DEBUG_PHOTOMETRY
         cerr << "Object at border .. " << xc << " " << yc << endl;
#endif
         ++nborder;
         continue;
      }
         
      MImage I = Image( Range(xc-wx, xc+wx), Range(yc-wy, yc+wy) );
        
      // check to see if data look reaonable ...
      if( !check_variance( I ) )
      {
         pd.ff_ = pd.dff_ = pd.fm_ = pd.dfm_ = pd.fsky_ = pd.dfsky_ = -99.;
         ++nbadregion;
#ifdef YODA_DEBUG_PHOTOMETRY
         cerr << "Object in bad region .. " << xc << " " << yc << endl;
#endif
         continue;
      }
         
      // if everything is ok, we can perform the photometry ...

      // (subarray has coordinates 1..2wx+1 => new center of aperture
      // in subimage is the fractional part of the coordinate + wx+1      
      float xcs = (pd.xc_-xc)+(wx+1);
      float ycs = (pd.yc_-yc)+(wy+1);

      // calculate sky
      double sky, sigma_sky;
      int nsky; 

      if( local_sky )
      {
         // the sky region is simply a difference of two elliptical masks ...
         MImage Sky = sky_e_mask( I, xcs, ycs, 
                                  aper+sky1+sky2, aper+sky1+sky2, 
                                  aper+sky1, aper+sky1, 0.0, nanval );
         // estimate a local sky value from an annulus around the object
         nsky = kappa_sigma_average( Sky, 2.5, 0.0f, &sky, &sigma_sky );
      }
      else
      {
         nsky = int(g.bggrid_*g.bggrid_*0.5);
         sky = bilin_interp( g.bg_, int(xc+0.5), int(yc+0.5), g.bggrid_ );
         sigma_sky=bilin_interp( g.rms_, int(xc+0.5), int(yc+0.5), g.bggrid_ );
      }
      
      // for the object, a smaller subimage is sufficient ...
      wx -= int(sky1+sky2);
      wy -= int(sky1+sky2);
      MImage A = Image( Range(xc-wx, xc+wx), Range(yc-wy, yc+wy) );
      xcs = (pd.xc_-xc)+(wx+1);
      ycs = (pd.yc_-yc)+(wy+1);
      
      // integrate flux-sky within subsampled aperture
      // for kron-like aperture
      float area=0.0;
      float flux=0.0;
      float skyerr2=0.0;
      float objerr2=0.0;
      float ferr=0.0;

      flux = integrate_soft_e_mask_fixed( A, sky, 
                                          xcs, ycs,
                                          aper,
                                          5, 
                                          &area );
      // error in sky estimate
      skyerr2 = area*area*sigma_sky*sigma_sky/nsky;
      // poisson error in flux + background noise
      objerr2 = (flux>0.0 ? flux/gain : 0.0) + 
         area*sigma_sky*sigma_sky;
      // total error
      ferr = sqrt( skyerr2 + objerr2 );
         
#ifdef YODA_DEBUG_PHOTOMETRY
      cerr << "   F:" << flux << " dF:" << ferr 
           << " S:" << sky << " dS:" << sqrt(skyerr2) << endl;
#endif
         
      pd.ff_ = flux;
      pd.dff_ = ferr;
      pd.fsky_ = sky;
      pd.dfsky_ = sigma_sky;
         
      if( flux > 0.0 )
      {
         pd.fm_  = -2.5*log10( pd.ff_ ) + magzero;
         pd.dfm_ =  2.5/M_LN10/pd.ff_*pd.dff_;
      }
      else
      {
         pd.fm_  = 99.;
         pd.dfm_ = 99;
      }
      
   }
   cerr << endl << "  " << nborder << " objects too close to or outside image border.\n";
   cerr << "  " << nbadregion << " objects in bad image regions.\n";
}


 

bool check_variance( const MImage& I )
{
   const int nx = I.maxIndex(1);
   const int ny = I.maxIndex(2);

   if( stddev( I(Range(1,nx/2), Range(1,ny/2)) ) < 1e-5 ||
       stddev( I(Range(1,nx/2), Range(ny/2,ny)) ) < 1e-5 ||
       stddev( I(Range(nx/2,nx),Range(1,ny/2))  ) < 1e-5 ||
       stddev( I(Range(nx/2,nx),Range(ny/2,ny)) ) < 1e-5 )
      return false;
   return true;
}


float integrate_soft_e_mask( const MImage& I, const float sky, 
                             float xc, float yc,
                             float a, float b, float pa,
                             const int subsample, float *area )
{
   const int nx = I.length(1)*subsample;
   const int ny = I.length(2)*subsample;
   const float ca = cos(pa*M_PI/180.);
   const float sa = sin(pa*M_PI/180.);
   a *= subsample;
   b *= subsample;
   xc *= subsample;
   yc *= subsample;
   
   float n=0.0, tmp=0.0, flux=0.0;
   const float val = 1.f/float(subsample*subsample);
   
   for( int j=1; j<=ny; ++j )
   {
      for( int i=1; i<=nx; ++i ) 
      {
         tmp = float(hypot( ( (i-xc)*ca+(j-yc)*sa )/a,
                            (-(i-xc)*sa+(j-yc)*ca )/b ) <= 1.) * val;
         n += tmp;
         flux += tmp * 
            ( I((i-1)/subsample+1,(j-1)/subsample+1) - sky );
      }
   }
   
   *area = n;
   
   return flux;
}


float integrate_soft_e_mask_fixed( const MImage& I, const float sky, 
                                   float xc, float yc,
                                   float ap, 
                                   const int subsample, float *area )
{
   const int nx = I.length(1)*subsample;
   const int ny = I.length(2)*subsample;
   xc *= subsample;
   yc *= subsample;   
   ap *= subsample;
   
   float n=0.0, tmp=0.0, flux=0.0;
   const float val = 1.f/float(subsample*subsample);
   
   for( int j=1; j<=ny; ++j )
   {
      for( int i=1; i<=nx; ++i ) 
      {
         tmp = float(hypot( i-xc, j-yc ) <= ap) * val;
         n += tmp;
         flux += tmp * 
            ( I((i-1)/subsample+1,(j-1)/subsample+1) - sky );
      }
   }
   
   *area = n;
   
   return flux;
}



MImage sky_e_mask( const MImage& I, const float xc, const float yc,
                   const float a1, const float b1,  
                   const float a2, const float b2,  
                   const float pa, const float nanval ) 
{
   MImage M( I.shape() );
   const float ca = cos(pa*M_PI/180.);
   const float sa = sin(pa*M_PI/180.);
   
   MImage::IndexIterator j = M.indexBegin();
   while( !j.done() )
   {
      if( feq(I(j()),nanval) )
      {
         M(j()) = 0;
      }
      else
      {
         M(j()) = ((hypot( ( (j(1)-xc)*ca+(j(2)-yc)*sa)/a1,
                           (-(j(1)-xc)*sa+(j(2)-yc)*ca)/b1 )<=1.) 
                 - (hypot( ( (j(1)-xc)*ca+(j(2)-yc)*sa)/a2,
                           (-(j(1)-xc)*sa+(j(2)-yc)*ca)/b2 )<=1.)) * I(j());
      }
      ++j;
   }
   
   return M;
}




// MImage soft_e_mask( const MImage& I, float xc, float yc,
//                     float a, float b, float pa,
//                     const int subsample, float *area )
// {
//    const int nx = I.length(1)*subsample;
//    const int ny = I.length(2)*subsample;
   
//    MImage M( I.shape() );
//    M = 0;

//    const float ca = cos(pa*M_PI/180.);
//    const float sa = sin(pa*M_PI/180.);
//    a *= subsample;
//    b *= subsample;
//    xc *= subsample;
//    yc *= subsample;
   
//    float n=0.0, tmp=0.0;
//    const float val = 1./float(subsample*subsample);
   
//    for( int j=1; j<=ny; ++j )
//    {
//       for( int i=1; i<=nx; ++i ) 
//       {
//          tmp = float(hypot( ( (i-xc)*ca+(j-yc)*sa )/a,
//                             (-(i-xc)*sa+(j-yc)*ca )/b ) <= 1.) * val;
//          n += tmp;
//          M( (i-1)/subsample+1, (j-1)/subsample+1 ) += tmp;
//       }
//    }
   
//    *area = n;
   
//    return M;
// }


// MImage soft_e_mask( const MImage& I, float xc, float yc,
//                     float a, float b, float pa,
//                     const int subsample, float *area )
// {
//    MImage M( I.length(1)*subsample, I.length(2)*subsample );
//    const float ca = cos(pa*M_PI/180.);
//    const float sa = sin(pa*M_PI/180.);
//    a *= subsample;
//    b *= subsample;
//    xc *= subsample;
//    yc *= subsample;
   
//    float n=0.0, tmp=0.0;
//    const float val = 1./float(subsample*subsample);
   
//    MImage::IndexIterator j = M.indexBegin();
//    while( !j.done() )
//    {
//       tmp = float(hypot( ( (j(1)-xc)*ca+(j(2)-yc)*sa )/a,
//                          (-(j(1)-xc)*sa+(j(2)-yc)*ca )/b ) <= 1.) * val;
      
//       M(j()) = tmp;
//       n += tmp;
//       ++j;
//    }

//    *area = n;
   
//    return ma_contract( M, subsample );
// }

// MImage soft_e_mask( const MImage& I, float xc, float yc,
//                     float a, float b, float pa,
//                     const int subsample )
// {
//    MImage M( I.length(1)*subsample, I.length(2)*subsample );
//    const float ca = cos(pa*M_PI/180.);
//    const float sa = sin(pa*M_PI/180.);
//    a *= subsample;
//    b *= subsample;
//    xc *= subsample;
//    yc *= subsample;
   

//    M = ( hypot( ( (indexPosFlt(M,1)-xc)*ca+(indexPosFlt(M,2)-yc)*sa)/a,
//                 (-(indexPosFlt(M,1)-xc)*sa+(indexPosFlt(M,2)-yc)*ca)/b )
//          <= 1.)
//       / float(subsample*subsample);
   
//    return ma_contract( M, subsample );
// }


// MImage annulus_e_mask( const MImage& I,  float xc,  float yc,
//                        float a1,  float b1,  
//                        float a2,  float b2,  
//                        float pa ) 
// {
//    MImage M( I.shape() );
//    const float ca = cos(pa*M_PI/180.);
//    const float sa = sin(pa*M_PI/180.);
   
//    M = (hypot( ( (indexPosFlt(M,1)-xc)*ca+(indexPosFlt(M,2)-yc)*sa)/a1,
//                (-(indexPosFlt(M,1)-xc)*sa+(indexPosFlt(M,2)-yc)*ca)/b1 )<=1.) 
//      - (hypot( ( (indexPosFlt(M,1)-xc)*ca+(indexPosFlt(M,2)-yc)*sa)/a2,
//                (-(indexPosFlt(M,1)-xc)*sa+(indexPosFlt(M,2)-yc)*ca)/b2 )<=1.);
   
//    return M;
// }



