/* 
 * classify.cpp
 * main routine for star/galaxy classification
 *
 * --------------------------------------------------------------------
 * $Id: classify.cpp 2 2002-08-19 17:00:29Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <list>

#include <ltl/marray.h>
#include <ltl/marray_io.h>
#include <ltl/fitsio.h>
#include <ltl/ascio.h>

//typedef MArray<float,2> MImage;


#include "yoda.h"
#include "globals.h"
#include "object.h"
#include "photdata.h"

extern Globals g;


list<PhotData*> init_classlist( const string& master );


void classify()
{
   MImage I;
   
   cerr << "Running resolution classifier ..."<<endl;	 

   if( !g.photometry_ )
   {
      // read image
      FitsIn infits( g.infile_ );
      g.Image_.makeReference( read_fits_image( infits ) );

      // set up GAIN
      if( g.gain_ == 0.0 )
      {
         try
         {
            g.gain_ = infits.getFloat( g.gainkey_ );
         }
         catch( FitsException& e )
	 {
	    cerr << "  Can't find GAIN keyword in FITS header!\n";
	    g.gain_ = 1.0;
	 }
      }
   }

   cerr << "  Using GAIN value of " << g.gain_ << endl;

   I.makeReference( g.Image_ );

   if( g.psftemplate_ != "" ) 
   {
      // read PSF template from file
      cerr << "  Reading PSF image from file '"+g.psftemplate_+"'\n";
      AscFile psffile( g.psftemplate_ );
      g.psfimage_.makeReference( psffile.readFloatColumns() );
   }
   else if( !g.psf_ )
   {
      cerr << "No PSF template specified!\n";
      cerr << "Specify either a PSF image, a catalog of objects in current\n"
           << "image to use, or use automatic selection of objects\n";
      exit(1);
   }
   
   // get target objects
   list<PhotData*> pl;
   
   if( g.photometry_ ) 
   {
      cerr << "  Using objects from detection/photometry pass ..." << endl;
      pl = g.photlist_;
   }
   else
   {
      pl = init_classlist( g.mastercat_ );
   }
   
   
   resolution_classifier( I, g.psfimage_, pl, g.gain_ );

   g.photlist_ = pl;
}



// get centers and sky from photometry catalog
//
list<PhotData*> init_classlist( const string& master ) 
{
   cerr << "  Reading object data from master catalog '"
	<<master<<"'" << endl;

   AscFile m( master );

   MVector f = m.readFloatColumn( 2 );
   MVector xc = m.readFloatColumn( 6 );
   MVector yc = m.readFloatColumn( 7 );
   MVector sky = m.readFloatColumn( 11 );

   int nmatch = 0;
   
   list<PhotData*> lp;
   for( int i=1; i<=xc.length(1); ++i ) 
   {
      list<Object*>::const_iterator ii = g.objlist_.begin();
      list<Object*>::const_iterator ie = g.objlist_.end();
      
      // find object in detected 
      for( ; ii != ie; ++ii )
      {
         Object& obj = **ii;
         if( hypot( xc(i)-obj.icx(), yc(i)-obj.icy() ) < 3 )
            break;
      }
      
      PhotData * pd = new PhotData();
         
      if( ii != ie )
      {  
         nmatch++;
         
         pd->f_ = f(i);
         pd->xc_ = xc(i);
         pd->yc_ = yc(i);
         pd->sky_ = sky(i);
      }
      else
      {
         pd->f_ = -1.;
         pd->xc_ = -1.;
         pd->yc_ = -1.;
         pd->sky_ = -.1;
      }
      
      lp.push_back( pd );

   }
   cerr << "  "<< nmatch << " objects recovered in image\n";
   
   return lp;   
}

