/* 
 * convolve.cpp
 * convolution with gaussian kernel (expression template version)
 *
 * --------------------------------------------------------------------
 * $Id: convolve.cpp 38 2004-10-06 21:40:32Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"
#include "yoda.h"
#undef LTL_RANGE_CHECKING   //causes enourmously long compile times

#include <ltl/marray.h>
#include <ltl/statistics.h>
#include <ltl/marray_io.h>

using ltl::Range;

// convolve an Image with a gaussian kernel
//
MImage convolve( const MImage& a, const float sigma, const float nan )
{
   const int lek=(sigma<2.0f ? 2 : 4);

   MImage res( a.shape() );
   res = nan;

   MImage k( Range(-lek,lek), Range(-lek,lek) );
  
   // setup normalized gaussian kernel
   for( int y=-lek; y<=lek; ++y )
      for( int x=-lek; x<=lek; ++x )
	 k(x,y) = exp(-(y*y+x*x)/(2.*sigma*sigma)) / (2*M_PI*sigma*sigma);

   // tweak normalization ...
   k /= sum(k);

   cerr.setf( ios::fixed, ios::floatfield );
   cerr.precision(2);
   cerr << "using gaussian kernel\n" << MArray<float,2>( k/k(0,0) ) << endl;
   cerr.setf( ios::floatfield, ios::fixed );
   cerr.precision(6);

   
   // now go for it
   const Range RX(a.minIndex(1)+lek, a.maxIndex(1)-lek);
   const Range RY(a.minIndex(2)+lek, a.maxIndex(2)-lek);

   res(RX,RY) = 0.0;
   for( int y=-lek; y<=lek; ++y )
      for( int x=-lek; x<=lek; ++x )
         res(RX,RY) += k(x,y)*a(RX-x,RY-y);
   
   res = feq(a,nan)*nan + fneq(a,nan)*res;
   return res;
}

/*
MImage convolve( const MImage& a, const float sigma, const float nan )
{
   const int lek=2;

   MImage res( a.shape() );
   res = nan;

   MImage k( Range(-lek,lek), Range(-lek,lek) );
  
   // setup normalized gaussian kernel
   for( int y=-lek; y<=lek; ++y )
      for( int x=-lek; x<=lek; ++x )
	 k(y,x) = exp(-(y*y+x*x)/(2.*sigma*sigma)) / (2*M_PI*sigma*sigma);

   // tweak normalization ...
   k /= sum(k);

   cerr.setf( ios::fixed, ios::floatfield );
   cerr.precision(2);
   cerr << "using gaussian kernel\n" << MArray<float,2>( k/k(0,0) ) << endl;
   cerr.setf( ios::floatfield, ios::fixed );
   cerr.precision(6);

   
   // now go for it
   const Range RX(a.minIndex(1)+lek, a.maxIndex(1)-lek);
   const Range RY(a.minIndex(2)+lek, a.maxIndex(2)-lek);

#if defined(__osf__) || defined(__xlC__)

   res(RX,RY) = 
        k(-2,-2)*a(RX-2,RY-2)   + k(-1,-2)*a(RX-1,RY-2) + 
      + k(0,-2)*a(RX,RY-2)    + k(1,-2)*a(RX+1,RY-2)  + k(+2,-2)*a(RX+2,RY-2);

   res(RX,RY) +=
        k(-2,-1)*a(RX-2,RY-1) + k(-1,-1)*a(RX-1,RY-1) + 
      + k(0,-1)*a(RX,RY-1)    + k(1,-1)*a(RX+1,RY-1)  + k(+2,-1)*a(RX+2,RY-1);
   
   res(RX,RY) +=
        k(-2,0)*a(RX-2,RY)    + k(-1,0)*a(RX-1,RY)    + 
      + k(0,0)*a(RX,RY)       + k(1,0)*a(RX+1,RY)     + k(+2,0)*a(RX+2,RY);
    
   res(RX,RY) +=
        k(-2,+1)*a(RX-2,RY+1) + k(-1,+1)*a(RX-1,RY+1) + 
      + k(0,+1)*a(RX,RY+1)    + k(1,+1)*a(RX+1,RY+1)  + k(+2,+1)*a(RX+2,RY+1);
   
   res(RX,RY) +=
        k(-2,+2)*a(RX-2,RY+2) + k(-1,+2)*a(RX-1,RY+2) + 
      + k(0,+2)*a(RX,RY+2)    + k(1,+2)*a(RX+1,RY+2)  + k(+2,+2)*a(RX+2,RY+2);
#else

   res(RX,RY) = 
      k(-2,-2)*a(RX-2,RY-2)   + k(-1,-2)*a(RX-1,RY-2) + 
      + k(0,-2)*a(RX,RY-2)    + k(1,-2)*a(RX+1,RY-2)  + k(+2,-2)*a(RX+2,RY-2)
    
      + k(-2,-1)*a(RX-2,RY-1) + k(-1,-1)*a(RX-1,RY-1) + 
      + k(0,-1)*a(RX,RY-1)    + k(1,-1)*a(RX+1,RY-1)  + k(+2,-1)*a(RX+2,RY-1)
    
      + k(-2,0)*a(RX-2,RY)    + k(-1,0)*a(RX-1,RY)    + 
      + k(0,0)*a(RX,RY)       + k(1,0)*a(RX+1,RY)     + k(+2,0)*a(RX+2,RY)
    
      + k(-2,+1)*a(RX-2,RY+1) + k(-1,+1)*a(RX-1,RY+1) + 
      + k(0,+1)*a(RX,RY+1)    + k(1,+1)*a(RX+1,RY+1)  + k(+2,+1)*a(RX+2,RY+1)
    
      + k(-2,+2)*a(RX-2,RY+2) + k(-1,+2)*a(RX-1,RY+2) + 
      + k(0,+2)*a(RX,RY+2)    + k(1,+2)*a(RX+1,RY+2)  + k(+2,+2)*a(RX+2,RY+2);

#endif
    
   res = feq(a,nan)*nan + fneq(a,nan)*res;
   return res;
}
*/
