/* 
 * detect.cpp
 * main routine for object detection pass
 *
 * --------------------------------------------------------------------
 * $Id: detect.cpp 34 2004-05-28 22:28:33Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <string>

#include "globals.h"
#include "yoda.h"
#include "object.h"

#include <ltl/marray.h>
#include <ltl/fitsio.h>

typedef MArray<float,2> MImage;
typedef MArray<char,2> MMask;

#include "interp.h"
#include "mathutil.h"

int int_ceil( const int p, const int q );
void dump_mask_image( MMask& M );

extern Globals g;

void detect()
{

   // read image
   FitsIn infits( g.infile_ );
   MImage I = read_fits_image( infits );
   
   // convolve image with Gaussian kernel and free original
   cerr << "Convolving image ... " << endl;
   MImage cI = convolve( I, g.seeing_/2.354820, g.nanvalue_ ); 
   I.free();
  
   const int nx = cI.length(1);
   const int ny = cI.length(2);  

   // create (small) background and rms images
   MImage  bg( int_ceil(nx,g.bggrid_), int_ceil(ny,g.bggrid_) );
   MImage rms( bg.shape() );
   MImage thr( bg.shape() );

   // compute background and rms maps
   background( cI, rms, bg );  
   g.rms_.makeReference( rms );
   g.bg_.makeReference( bg );
   
   // threshold using interpolated background and rms
   cerr << "Thresholding at " << g.thresh_ << " x local noise ..." << endl;
   thr = bg + g.thresh_ * rms;
   const float nanvalue = g.nanvalue_;
   MMask M = ( fneq(cI,nanvalue) ) &&
      ( cI >= interp2d( indexPos(cI,1), indexPos(cI,2), 
			thr, g.bggrid_, g.bggrid_ ) );
   cI.free();

   // reread image
   MImage I2 = read_fits_image( infits, true );

   // subtract background from image
   I2 -= interp2d( indexPos(cI,1), indexPos(cI,2), 
		   bg, g.bggrid_, g.bggrid_ );
   // now we have to assemble the flagged pixels into objects
   cerr << "Assembling objects ..." << endl;
   cerr << "  Minimum number of consecutive pixels above threshold: "
        << g.npixmin_ << endl;
   list<Object*> lst = assemble( M, I2, rms, 3, false );

   // dump mask image 
   if( g.bg_dump_mask_ )
   {
      M = 0;
      for( list<Object*>::iterator oi = lst.begin(); oi!=lst.end(); ++oi )
         (*oi)->markpixels( M );
      dump_mask_image( M );
   }
   
   // set the (initial) detection isophote
   for( list<Object*>::iterator oi = lst.begin(); oi!=lst.end(); ++oi )
      (*oi)->setthresh( g.thresh_*bilin_interp( rms, 
						(*oi)->maxx(), (*oi)->maxy(), 
						g.bggrid_) );

   // splitting is next
   cerr << "Splitting objects ..." << endl;
   list<Object*> lst2 = split( lst );
   delete_list( lst );

   // evaluate objects
   cerr << "Evaluating objects ..." << endl;
   g.objlist_ = evaluate( lst2, I2 );
   g.Image_.makeReference( I2 );
}


void determine_sky()
{
   // read image
   FitsIn infits( g.infile_ );
   MImage I = read_fits_image( infits );
   
   const int nx = I.length(1);
   const int ny = I.length(2);  

   // create (small) background and rms images
   MImage  bg( int_ceil(nx,g.bggrid_), int_ceil(ny,g.bggrid_) );
   MImage rms( bg.shape() );
   MImage thr( bg.shape() );

   // compute background and rms maps
   background( I, rms, bg );  
   g.rms_.makeReference( rms );
   g.bg_.makeReference( bg );   
}



void dump_mask_image( MMask& M )
{
   FitsOut dfits( prepend_filename( g.infile_, "detect_" ) );
   dfits.addValueCard( "OBJECT  ", "Detection image", "Created by YODA" );
   dfits.addValueCard( "ORIG_IMG", g.infile_, "Original image file" );
   dfits.setOrigin( "YODA" );
   dfits << M;
}

// return ceil p/q for integer division
// i.e. p/q+1 if p/q has a rest, else p/q
int int_ceil( const int p, const int q )
{
   return p/q + ((p%q)!=0);
}
