/*
 * interp.cpp
 * Bilinear interpolation
 *
 * --------------------------------------------------------------------
 * $Id: interp.cpp 2 2002-08-19 17:00:29Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <cmath>
#include <ltl/marray.h>


using namespace ltl;
typedef MArray<float,2> MImage;

// f(x,y) = a1 + a2*x + a3*y + a4*x*y
// where the coefficients a1...a4 are determined from f(x0,y0) ... f(x1,y1),
// with (x0,y0), (x1,y0), (x0,y1), and (x1,y1) being the 4 neighbouring
// sampling points

#define CLAMP(x,low,high) (((x)>(high)) ? (high) : (((x)<(low)) ? (low) : (x)))


float bilin_interp( const MImage& I, const int x, const int y, 
		    const int mesh )
{
   // calculate mesh cell in sample_ array
   int xF = (x+mesh/2)/mesh;
   int yF = (y+mesh/2)/mesh;

   xF = CLAMP(xF, 1, I.maxIndex(1)-1);
   yF = CLAMP(yF, 1, I.maxIndex(2)-1);

   const float F00 = I(xF   , yF);
   const float F01 = I(xF   , yF+1);
   const float F10 = I(xF+1 , yF);
   const float F11 = I(xF+1 , yF+1);
  
   const float X0 = xF*mesh - mesh/2;
   const float X1 = X0 + mesh;
   const float Y0 = yF*mesh - mesh/2;
   const float Y1 = Y0 + mesh;

   const float D = (X1*Y1 - X1*Y0 - X0*Y1 + X0*Y0);
  
   const float a1 = (F00*X1*Y1 - F01*X1*Y0 - F10*X0*Y1 + F11*X0*Y0)/mesh/mesh;
   const float a2 = (F10*Y1 - F00*Y1 - F11*Y0 + F01*Y0) / D;
   const float a3 = (F01*X1 + F10*X0 - F11*X0 - F00*X1) / D;
   const float a4 = (F11 - F01 + F00 - F10) / mesh / mesh;

   // finally return the interpolant
   return a1 + a2*x + a3*y + a4*x*y;
}



//  float interp2d( const MImage& I, const float x, const float y )
//  {
//    int x1 = int(floor(x));
//    int y1 = int(floor(y)); 

//    x1 = CLAMP(x1, I.minIndex(1), I.maxIndex(1)-1);
//    y1 = CLAMP(x1, I.minIndex(2), I.maxIndex(2)-1);

//    const int y2 = y1+1;
//    const int x2 = x1+1;

//    const float t = (x-x1);
//    const float u = (y-y1);

//    return (1.0f-t)*(1.0f-u)*I(x1,y1) + 
//           t*(1.0f-u)*I(x2,y1) + 
//           t*u*I(x2,y2) + 
//           (1.0f-t)*I(x1,y2);
//  }



float interp2d( const MImage& I, const float x, const float y )
{
   // calculate mesh cell in sample_ array
   int X0 = int(floor(x));
   int Y0 = int(floor(y));

   X0 = CLAMP(X0, I.minIndex(1), I.maxIndex(1)-1);
   Y0 = CLAMP(Y0, I.minIndex(2), I.maxIndex(2)-1);
   
   int X1 = X0 + 1;
   int Y1 = Y0 + 1;

   const float F00 = I(X0, Y0);
   const float F01 = I(X0, Y1);
   const float F10 = I(X1, Y0);
   const float F11 = I(X1, Y1);

   const float a1 = (F00);
   const float a2 = (F10 - F00); 
   const float a3 = (F01 - F00); 
   const float a4 = (F11 - F01 + F00 - F10); 

   const float xx = x-X0;
   const float yy = y-Y0;
   // finally return the interpolant
   return a1 + a2*xx + a3*yy + a4*xx*yy;
}



