/* 
 * io.cpp
 * misc io routines
 *
 * --------------------------------------------------------------------
 * $Id: io.cpp 6 2002-10-29 17:17:10Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <fstream>
#include <iomanip>
#include <string>

#include <ltl/marray.h>
#include <ltl/fitsio.h>
#include <ltl/util/utdate.h>

using ltl::FitsException;

#include "yoda.h"
#include "globals.h"
#include "object.h"
#include "photdata.h"


extern Globals g;
extern char* version_str;

void write_cmd_line( ostream& os );


string prepend_filename( string path, const string& prefix )
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}


MImage read_fits_image( FitsIn& infits, const bool quiet )
{
   if( !quiet ) 
   {
      //infits.describeSelf( cerr ); 
      try 
      {
         string object = infits.getString( "OBJECT  " );
         cerr << "OBJECT=" << object << endl << endl;
      }
      catch( FitsException& e )
      {
      }
   }

   // get Image 
   MImage A;
   infits >> A;
   infits.freeData();
   
   return A;
}


void write_master_catalog( list<Object*> l, string& filename )
{
   // write catalog file
   int n=1;
   ostream* outstream = &cout;
   ofstream outfile;

   if( filename == "<Image>.cat" )
      filename = g.infile_.substr( 0, g.infile_.length()-4 ) + "cat"; 

   if( filename!="stdout" ) 
   {
      cerr << "Writing master catalog to '"<<filename<<"'.\n";
      outfile.open( filename.c_str() );
      if( !outfile.good() ) 
      {
	 cerr << "[yoda] can't create catalog file '"<<filename<<"'!\n";
	 exit(-1);
      }
      outstream = &outfile;
   }
  
   write_cmd_line( *outstream );
   
   *outstream << "# IMAGE : '" << g.infile_ << "'" << endl;
   *outstream << 
      "# ID     X         Y      MAG     IXX     IYY     IXY     ASY    FWHM    IR1     A      B       PA\n";
  
   for( list<Object*>::iterator i=l.begin(); i!=l.end(); ++i, ++n ) 
   {
      Object &obj = **i; 
      *outstream << setw(5) << n;
      obj.writecat( *outstream );
      *outstream << endl;
   }

}

void write_photometry_catalog( list<PhotData*> l, string& filename )
{
   // write catalog file
   int n=1;
   ostream* outstream = &cout;
   ofstream outfile;

   if( filename == "<Image>.p" )
      filename = g.infile_.substr( 0, g.infile_.length()-4 ) + "p"; 

   if( filename!="stdout" ) 
   {
      cerr << "Writing photometry catalog to '"<<filename<<"'.\n";
      outfile.open( filename.c_str() );
      if( !outfile.good() ) 
      {
	 cerr << "[yoda] can't create catalog file '"<<filename<<"'!\n";
	 exit(-1);
      }
      outstream = &outfile;
   }
  
   write_cmd_line( *outstream );

   *outstream << "# IMAGE : '" << g.infile_ << "'";
   if( g.mastercat_ != "" )
      *outstream << ",  MASTER-CAT : '" << g.mastercat_ << "'";
   if( g.coordscat_ != "" )
      *outstream << ",  COORDS-CAT : '" << g.coordscat_ << "'";
   *outstream << endl;
   
   *outstream << 
      "#  ID  K-FLUX      FLUXERR     K-MAG    MAGERR X       Y       A_AP   B_AP    PA        SKY      SKYERR  FIX-FLUX    FLUXERR     FIX-MAG MAGERR FIXRADIUS   FIXSKY  SKYERR" << endl;

   for( list<PhotData*>::iterator i=l.begin(); i!=l.end(); ++i, ++n ) 
   {
      PhotData &obj = **i; 
      *outstream << setw(5) << n;
      obj.writecat( *outstream );
      *outstream << endl;
   }

}

void write_psffile( const MImage& psf )
{
   ofstream psffile;
   if( g.psffile_ == "<Image>.psf" )
      g.psffile_ = g.infile_.substr( 0, g.infile_.length()-4 ) + "psf"; 

   cerr << "Writing PSF image to '"<<g.psffile_<<"'.\n";

   psffile.open( g.psffile_.c_str() );
   if( !psffile.good() ) 
   {
      cerr << "[yoda] can't create PSF image file '"<<g.psffile_<<"'!\n";
      exit(-1);
   }
   
   write_cmd_line( psffile );   
   psffile << "# PSF for '"<< g.infile_ << "'\n";
   
   for( int j=psf.minIndex(2); j <= psf.maxIndex(2); ++j ) 
   {
      for( int i=psf.minIndex(1); i <= psf.maxIndex(1); ++i ) 
      {
	 psffile << setw(8) << setprecision(3) << psf(i,j) << " ";
      }
      psffile << endl;
   }
   psffile.close();
}


void write_classify_catalog( list<PhotData*> l, string& filename )
{
   // write catalog file
   int n=1;
   ostream* outstream = &cout;
   ofstream outfile;

   if( filename == "<Image>.class" )
      filename = g.infile_.substr( 0, g.infile_.length()-4 ) + "class"; 

   if( filename!="stdout" ) 
   {
      cerr << "Writing classifier catalog to '"<<filename<<"'.\n";
      outfile.open( filename.c_str() );
      if( !outfile.good() ) 
      {
	 cerr << "[yoda] can't create catalog file '"<<filename<<"'!\n";
	 exit(-1);
      }
      outstream = &outfile;
   }
  
   write_cmd_line( *outstream );

   *outstream << "# IMAGE : '" << g.infile_ << "'";
   if( g.mastercat_ != "" )
      *outstream << ",  MASTER-CAT : '" << g.mastercat_ << "'";
   if( g.coordscat_ != "" )
      *outstream << ",  COORDS-CAT : '" << g.coordscat_ << "'";
   *outstream << endl;
   
   *outstream << 
"# ID ALPHA  BETA  LUM\n";

   outstream->setf( ios::fixed, ios::floatfield );

   for( list<PhotData*>::iterator i=l.begin(); i!=l.end(); ++i, ++n ) 
   {
      PhotData &obj = **i; 
      *outstream << setw(5) << n << "  ";
      *outstream << setw(5) << setprecision(3) << obj.alpha_ << " ";   
      *outstream << setw(5) << setprecision(3) << obj.beta_  << " ";  
      *outstream << setw(5) << setprecision(3) << obj.lum_  << endl;  
   }

   outstream->setf( ios::floatfield, ios::fixed );
   
}




void write_cmd_line( ostream& os )
{
   int n = 0;
   os << "# CALL : ";
   while( n<g.argc_ )
   {
      os << g.argv_[n++] << " ";
   }
   os << endl;
   os << "# THIS OUTPUT WAS CREATED ON : " << util::UTDate().toString(); 
   os << "  by YODA version " << version_str << endl;
}




/*
void write_pixellist( list<Object*> *l, const char* filename )
{
// write pixel list file

int n=1;

ofstream pixfile;
cerr << "Writing pixel list file to '"<<filename<<"'.\n";
pixfile.open( filename );
if( !pixfile.good() ) {
cerr << "[yoda] can't create pixel list file '"<<filename<<"'!\n";
exit(-1);
}
  
for( list<Object*>::iterator i=l->begin(); i!=l->end(); ++i, ++n ) {
Object &obj = **i; 
pixfile << n << "  " << obj;
}    

pixfile.close();
}


*/



