/* -*- C++ -*-
 *
 * object.cpp
 * Implementation of class object
 *
 * --------------------------------------------------------------------
 * $Id: object.cpp 22 2003-09-22 18:46:51Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */


#include "config.h"

#include "globals.h"

//#include "yoda.h"
#include "object.h"

#include <iostream>
#include <iomanip>
#include <cmath>

using namespace std;

extern Globals g;



Object::~Object() 
{
   //pixels_.clear();
}


void Object::addpixel( const int x, const int y, const float v ) 
{
   // calculate extent in x, y
   if( pixnum_ == 0 ) 
   {
      // first pixel
      x1_ = x2_ = x;
      y1_ = y2_ = y;
      mx_ = x;
      my_ = y;
      mv_ = v;
   } 
   else 
   {
      // all others
      if( x < x1_ ) x1_ = x;
      if( y < y1_ ) y1_ = y;
      if( x > x2_ ) x2_ = x;
      if( y > y2_ ) y2_ = y;
      if( v > mv_ ) {
	 mx_ = x;
	 my_ = y;
	 mv_ = v;
      } 
   }
   // add pixel to list
   pixnum_++;
   pixels_.push_front ( Pixel( x, y, v ) );
}


void Object::setpixels( MImage& image, const float v )
{
   for( Pixiter i=pixels_.begin(); i!=pixels_.end(); ++i ) 
   {
      const Pixel &p=*i;
      image( p.x(), p.y() ) = v;
   }
}


void Object::markpixels( MMask& image )
{
   for( Pixiter i=pixels_.begin(); i!=pixels_.end(); ++i ) 
   {
      const Pixel &p=*i;
      image( p.x(), p.y() ) = 1;
   }
}


void Object::drawpixels( MImage& image )
{
   for( Pixiter i=pixels_.begin(); i!=pixels_.end(); ++i ) 
   {
      const Pixel &p=*i;
      image( p.x(), p.y() ) = p.v();
   }
}


float Object::pixelmedian()
{
   pixels_.sort();
   Pixiter median=pixels_.begin();
   advance(median, pixnum()/2);
   return median->v();
}


bool Object::evaluate( const MImage& image )
{
   evalgood_ = false;
   float isum = 0.0f;
   isomag_=0.0f;
   lum_ = 0.0f;
   icx_ = icy_ = ixx_ = iyy_ = ixy_ = ir1_ = 0.0f;

   // first, intensity weighted centroids, and isphotal magnitude
   for( Pixiter i=pixels_.begin(); i!=pixels_.end(); ++i ) 
   {
      const Pixel &p = *i;
      const float v = image( p.x(), p.y() );
      isum += v;
      icx_ += p.x() * v;
      icy_ += p.y() * v;
      isomag_ += v;
   }
   icx_ /= isum;
   icy_ /= isum;
  
   lum_ = isum;
  
   if( isomag_ > 0.0 )
      isomag_ = -2.5*log10(isomag_) + g.magzero_;
   else
      return false;

   // intensity weighted second moments
   for( Pixiter i=pixels_.begin(); i!=pixels_.end(); ++i ) 
   {
      const Pixel &p=*i;
      const float dy = p.y() - icy_;
      const float dx = p.x() - icx_;
      const float v = image( p.x(), p.y() );
      ixx_ += dx * dx * v;
      iyy_ += dy * dy * v;
      ixy_ += dx * dy * v;
   }
   ixx_ /= isum;
   iyy_ /= isum;
   ixy_ /= isum;
   fwhm_ = 2.0f*sqrt(M_LN2*(ixx_+iyy_));

   // check second moments
   float lall = 0.5*(ixx_+iyy_) -
      std::sqrt( 0.25*(ixx_-iyy_)*(ixx_-iyy_) + ixy_*ixy_ );
   if( lall < 0.0 )
      return false;

   // asymmetry
   asym_ = sqrt( (iyy_-ixx_)*(iyy_-ixx_) + 4.*ixy_*ixy_ ) / (ixx_+iyy_);

   // now check everything is reasonable
   float dx=x2_-x1_;
   float dy=y2_-y1_;
   if( std::sqrt( dx*dx + dy*dy ) > (float)g.maxsize_ )
      return false;

   calc_lum();

   calc_aper();

   // first moment
   calc_ir1( image );

   evalgood_=true;
   return true;
}


void Object::calc_lum()
{
   int n = pixnum() >= 9 ? 9 : pixnum();
   pixels_.sort();
   Pixiter i=pixels_.begin();
   advance(i, pixnum()-n);
  
   corelum_ = 0.0f;
   for( ; i!=pixels_.end(); ++i )
      corelum_ += (*i).v();
   corelum_ /= n;

   Pixiter i2=pixels_.begin();
   advance(i2, n);
   isolum_ = 0.0f;
   for( i=pixels_.begin(); i!=i2; ++i )
      isolum_ += (*i).v();
   isolum_ /= n;
}


void Object::calc_aper()
{
   float a = 0.5*(ixx_+iyy_);
   float c = sqrt( 0.25*(ixx_-iyy_)*(ixx_-iyy_) + ixy_*ixy_ );
  
   // semi-major and semi-minor axes
   a_  = sqrt( a+c );
   b_  = a-c > 0 ? sqrt(a-c) : a_;

   // the kron-like aperture a' and b' are then 
   // (with e=a/b) : cd /k*ir1*e and k*ir1/e

   pa_ = ixx_-iyy_ != 0.0 ? 0.5*atan2( (float)2.0*ixy_, float(ixx_-iyy_) ) : M_PI/4.;

   // position angle in degree
   pa_ *= 180./M_PI;
}



void Object::calc_ir1( const MImage& image )
{
   float cxx, cyy, cxy, t;
   
   t = ixx_*iyy_ - ixy_*ixy_;
   if( t < 0.00694 )
      t = (ixx_+0.084)*(iyy_+0.084) - ixy_*ixy_;
   cxx = ixx_/t;
   cyy = iyy_/t;
   cxy = -2.0*ixy_/t;
   
   // calculate ir1 within detected area 
   for( Pixiter i=pixels_.begin(); i!=pixels_.end(); ++i ) 
   {
      const Pixel &p=*i;
      const float y = p.y() - icy_;
      const float x = p.x() - icx_;
      const float v = image( p.x(), p.y() );
      ir1_ += v*a_*sqrt(cxx*x*x + cyy*y*y + cxy*x*y);      
   }
   ir1_ /= lum_;
}


void Object::writecat( ostream& os )
{
   os.setf( ios::fixed, ios::floatfield );
   os << setw(9) << setprecision(2) << icx_    << " ";    // 2
   os << setw(8) << setprecision(2) << icy_    << " ";    // 3
   os << setw(7) << setprecision(2) << isomag_ << " "; 	 // 4
   os << setw(7) << setprecision(3) << ixx_    << " ";    // 5
   os << setw(7) << setprecision(3) << iyy_    << " ";    // 6
   os << setw(7) << setprecision(3) << ixy_    << " ";    // 7
   os << setw(7) << setprecision(3) << asym_   << " ";  	 // 8
   os << setw(6) << setprecision(3) << fwhm_   << " ";  	 // 9
   os << setw(7) << setprecision(3) << ir1_    << " ";    // 10      
   os << setw(7) << setprecision(3) << a_      << " ";    // 11
   os << setw(7) << setprecision(3) << b_      << " ";    // 12
   os << setw(7) << setprecision(3) << pa_;               // 13
   os.setf( ios::floatfield, ios::fixed );
}



ostream& operator<<( ostream& os, Object& obj )
{
   os << "{" << endl;
   for( Object::Pixiter i=obj.pixels_.begin(); i!=obj.pixels_.end(); ++i ) 
   {
      const Pixel &p=*i;
      os << p.x() << " " << p.y() << " " << p.v() << endl;
   }
   os << "}" << endl;
   return os;
}

istream& operator>>( istream& is, Object& obj )
{
   char c;

   // advance past '{'
   is >> c;
   if( c != '{' ) 
   {
      cerr << "[Object::read] : '{' expected\n";
      is.clear( ios::badbit ); // SET bad state (stdlib is wierd here ...
      return is;
   }
  
   is >> c;
   // read pixels
   while( c != '}' ) 
   {
      is.putback(c);
      int x, y;
      float v;
      is >> x >> y >> v;
      obj.addpixel( x, y, v );
      is >> c;
   }
   return is;
}
