/* -*- C++ -*-
 *
 * object.h
 * Declaration of class object
 *
 * --------------------------------------------------------------------
 * $Id: object.h 42 2014-09-24 22:58:44Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#ifndef __YODA_OBJECT__
#define __YODA_OBJECT__

#include <ltl/marray.h>

using  namespace ltl;

typedef MArray<float,2> MImage;
typedef MArray<char,2> MMask;


class Object;

class Pixel
{
   public:
      friend class Object;

      Pixel( const int x, const int y, const float v )
	 : x_(x), y_(y), v_(v)
      { }
  
      Pixel( const Pixel &other )
	 : x_(other.x_), y_(other.y_), v_(other.v_)
      { }
  
      bool operator<( const Pixel& other ) const
      { return v() < other.v(); }

      int x() const
      { return x_; }

      int y() const
      { return y_; }
  
      float v() const
      { return v_; }
    
   protected:
      int x_, y_;
      float v_;
};




class Object
{
   public:
      typedef list<Pixel> Pixlist;
      typedef Pixlist::const_iterator Pixiter;

      friend ostream& operator<<( ostream& os, Object& obj );
      friend istream& operator>>( istream& is, Object& obj );
  
      Object() 
	 : evalgood_(0), split_(false), pixnum_(0), mx_(0), my_(0), mv_(0),
	   th_(0), x1_(0), y1_(0), x2_(0), y2_(0),
	   icx_(0), icy_(0), ixx_(0), iyy_(0), ixy_(0), asym_(0), ir1_(0), 
	   fwhm_(0),               
	   a_(0), b_(0), pa_(0),               
	   lum_(0), isomag_(0),             
	   isolum_(0), corelum_(0)         
      { };

      Object( const Object& other ) 
	 :  pixels_(other.pixels_),
	    evalgood_(other.evalgood_), 
	    split_(other.split_), 
	    pixnum_(other.pixnum_), 
	    mx_(other.mx_), my_(other.my_), mv_(other.mv_),
	    th_(other.th_), 
	    x1_(other.x1_), y1_(other.y1_), x2_(other.x2_), y2_(other.y2_),
	    icx_(other.icx_), icy_(other.icy_), 
	    ixx_(other.ixx_), iyy_(other.iyy_), ixy_(other.ixy_), 
	    asym_(other.asym_), ir1_(other.ir1_), fwhm_(other.fwhm_),      
	    a_(other.a_), b_(other.b_), pa_(other.pa_),               
	    lum_(other.lum_), isomag_(other.isomag_),             
	    isolum_(other.isolum_), corelum_(other.corelum_)         
      { };

      ~Object();

      bool operator==( const Object& other )
      { return maxx()==other.maxx() && maxy()==other.maxy(); }

      bool operator!=( const Object& other )
      { return maxx()!=other.maxx() || maxy()!=other.maxy(); }

      void addpixel( const int x, const int y, const float v );

      int pixnum() const
      { return pixnum_; }

      bool evaluate( const MImage& image );

      bool evalgood() const
      { return evalgood_; }

      void drawpixels( MImage& image );
      void markpixels( MMask& image );
      void setpixels( MImage& image, const float v );

      void writecat( ostream& os );

      float pixelmedian();

      void setthresh( float t )
      { th_ = t; }
      void setsplit()
      { split_ = true; }
      bool getsplit() const
      { return split_; }
  
      float thresh() const
      { return th_; }

      int xmin() const
      { return x1_; }
      int xmax() const
      { return x2_; }
      int ymin() const
      { return y1_; }
      int ymax() const
      { return y2_; }

      int maxx() const
      { return mx_; }
      int maxy() const
      { return my_; }
      float maxv() const
      { return mv_; }

      float icx() const
      { return icx_; }
      float icy() const
      { return icy_; }
      float ir1() const
      { return ir1_; }
      float a() const
      { return a_; }
      float b() const
      { return b_; }
      float pa() const
      { return pa_; }

      float fwhm() const
      { return fwhm_; }

      float asym() const
      { return asym_; }

      float lum() const
      { return lum_; }
      float isolum() const
      { return isolum_; }
      float corelum() const
      { return corelum_; }

   protected:
      void init();
      void calc_lum();
      void calc_aper();
      void calc_ir1( const MImage& image );

      Pixlist pixels_;

      bool evalgood_;
      bool split_;
      int  pixnum_;

      int mx_, my_;  // coordinates of maximum intensity
      float mv_;     // max intensity
      float th_;     // local threshhold

      int x1_, y1_, x2_, y2_;           // minimal and maximal extents
      float icx_, icy_;                 // first moments 
      float ixx_, iyy_, ixy_, asym_;    // second moments
      float ir1_, fwhm_;                // intens. weighted radius
      float a_, b_, pa_;                // semi-major and minor axes, pos-angel
      float lum_, isomag_;              // luminosity and isophotal magnitude
      float isolum_, corelum_;          // lum at detection isophote, core lum
};

#endif
