/* 
 * photometry.cpp
 * main routine for aperture photometry pass
 *
 * --------------------------------------------------------------------
 * $Id: photometry.cpp 37 2004-10-06 21:12:28Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <iostream>
#include <string>
#include <list>

#include "globals.h"
#include "yoda.h"
#include "object.h"
#include "photdata.h"

#include <ltl/marray.h>
#include <ltl/marray_io.h>
#include <ltl/fitsio.h>
#include <ltl/ascio.h>

typedef MArray<float,1> MVector;


extern Globals g;

list<PhotData*> init_photlist( const list<Object*>& l );
list<PhotData*> init_photlist( const string& master );
list<PhotData*> init_photlist( const string& coords, bool dummy );
list<PhotData*> init_photlist( const string& master, const string& coords );

bool need_to_transform();
void iraf2matrix();
void linear_transform_ap( MVector& a, MVector& b, MVector& pa );

void photometry( )
{
   cerr << "Performing aperture photometry ..."<<endl;
	 
   // read image
   FitsIn infits( g.infile_ );
   MImage I = read_fits_image( infits );
   g.Image_.makeReference( I );
   
   // We need to arrange ourselves with the parameter situation first
   //   o  neither mastercat nor coordscat given : use detections
   //   o  mastercat given, no coordscat : everything from mastercat
   //   o  mastercat and coordscat given : aperture sizes from mastercat,
   //      centers from coordscat
   //   o  no mastercat, only coordscat : ERROR, unless we don't need
   //      to perform kron-photometry
   list<PhotData*> pl;

   if( g.mastercat_ == "" )          // no master cat
   {
      if( g.coordscat_ == "" )       // no coords cat
      {
	 // get everything from detected objects
	 pl = init_photlist( g.objlist_ );
      } 
      else                           // no master, only coords
      {
         if( g.no_kron_ap_ )         // can't do kron apertures
         {
            pl = init_photlist( g.coordscat_, true );
         }
         else
         {
            // ERROR
            cerr << "[FATAL] can't do kron-apertures without master catalog\n";
            cerr << "ABORTING ...\n\n" ;
            exit(1);
         }
      }
   }
   else                               // master cat
   {
      if( g.coordscat_ == "" )        // no coords cat
      {
	 // get everything from mastercat
	 pl = init_photlist( g.mastercat_ );      
      }
      else                            // master cat and coords cat
      {
	 // get apertures from mastercat, coordinates from coordscat
	 pl = init_photlist( g.mastercat_, g.coordscat_ );
      }    
   }

   
   // set up GAIN
   if( g.gain_ == 0.0 )
   {
      try
      {
         g.gain_ = infits.getFloat( g.gainkey_ );
      }
      catch( FitsException& e )
      {
         cerr << "  Can't find GAIN keyword in FITS header!\n";
         g.gain_ = 1.0;
      }
   }

   cerr << "  Using GAIN value of " << g.gain_ << endl;
   
   if( g.fixed_aper_ != 0.0 ) 
   {
      if( !g.no_kron_ap_ )
      cerr << "  Performing photometry using " << g.fixed_aper_ << " pixel diameter apertures." << endl;
   }
   
   // ok, now the photometry itself ...
   // results are written into the PhotDat objects which then 
   // also contain the USED APERTURE SIZES
   const bool local_sky = (g.skyann_[0] > 0.5) && (g.skyann_[1]>0.5);
   if( local_sky )
      cerr << "  Using sky annulus: " << g.skyann_[0] << "," << g.skyann_[1] << endl;
   else
      cerr << "  Using sky from interpolated background " << endl;

   if( !g.no_kron_ap_ ) 
   {
      cerr << "  Performing photometry using kron-like apertures," << endl;
      cerr << "  using a Kron-factor of "<< g.kronfactor_ << endl;
      
      kron_aperture_phot( I, pl, g.kronfactor_, g.minaperture_, 
                          g.skyann_[0], g.skyann_[1],
                          g.gain_, g.magzero_, g.nanvalue_ );
   }
   if( g.fixed_aper_ != 0.0 )
   {
      cerr << "  Performing photometry using " << g.fixed_aper_ << 
         " pixel (diameter) apertures." << endl;
      
      fixed_aperture_phot( I, pl, g.fixed_aper_/2.,
                           g.skyann_[0], g.skyann_[1],
                           g.gain_, g.magzero_, g.nanvalue_ );     
   }
   g.photlist_ = pl;
}




// get apertures and positions from detected objects
//
list<PhotData*> init_photlist( const list<Object*>& l )
{
   cerr << "  Reading positions/apertures from detected sources." << endl;
   
   list<PhotData*> lp;
   list<Object*>::const_iterator i=l.begin();
   for( ; i!=l.end(); ++i ) 
   {
      const Object& obj = **i;
      lp.push_back( new PhotData( obj.icx(), obj.icy(), 
				  obj.a(), obj.b(), obj.pa(), obj.ir1(),
                                  g.fixed_aper_ ) );
   }
   return lp;
}


// get apertures and positions from master catalog
//
list<PhotData*> init_photlist( const string& master ) 

{
   cerr << "  Reading positions and apertures from master catalog '"
	<<master<<"'" << endl;

   AscFile m( master );

   MVector xc = m.readFloatColumn( 2 );
   MVector yc = m.readFloatColumn( 3 );

   MVector ir1= m.readFloatColumn( 10 );
   MVector a  = m.readFloatColumn( 11 );
   MVector b  = m.readFloatColumn( 12 );
   MVector pa = m.readFloatColumn( 13 );

   list<PhotData*> lp;
   for( int i=1; i<=xc.length(1); ++i ) 
   {
      lp.push_back( new PhotData( xc(i), yc(i), a(i), b(i), pa(i), ir1(i),
                                  g.fixed_aper_/2. ) );
   }
   return lp;   
}


// get positions from coord catalog, no apertures
// (we're doing only fixed aperture stuff ...)
//
list<PhotData*> init_photlist( const string& coords, bool dummy ) 

{
   const int xcol = int(g.coordscolumns_[0]);
   const int ycol = int(g.coordscolumns_[1]);

   cerr << "  Reading positions from coordinates catalog '"
	<<coords<<"', no kron-apertures" << endl;
   cerr << "  using columns " << xcol << " and " << ycol << endl;

   AscFile m( coords );

   MVector xc = m.readFloatColumn( xcol );
   MVector yc = m.readFloatColumn( ycol );

   list<PhotData*> lp;
   for( int i=1; i<=xc.length(1); ++i ) 
   {
      lp.push_back( new PhotData( xc(i), yc(i), 
                                  0.0, 0.0, 0.0, 0.0,
                                  g.fixed_aper_/2. ) );
   }
   return lp;   
}


// get apertures from master catalog and positions from coords file
// transform apertures using liner transform given on command line
//
list<PhotData*> init_photlist( const string& master, const string& coords )
{
   const int xcol = int(g.coordscolumns_[0]);
   const int ycol = int(g.coordscolumns_[1]);

   cerr << "  Reading positions from coordinates catalog '"
	<<coords<<"'" << endl;
   cerr << "  using columns " << xcol << " and " << ycol << endl;
   cerr << "  Reading apertures from master catalog '"
	<<master<<"'" << endl;

   AscFile c( coords );
   AscFile m( master );
   
   MVector xc = c.readFloatColumn( xcol );
   MVector yc = c.readFloatColumn( ycol );
   MVector ir1= m.readFloatColumn( 10 );
   MVector a  = m.readFloatColumn( 11 );
   MVector b  = m.readFloatColumn( 12 );
   MVector pa = m.readFloatColumn( 13 );

   // have to transform apertures ...
   if( need_to_transform() ) 
   {
      if( g.iraftrans_ )
	 iraf2matrix();   
      linear_transform_ap( a, b, pa );
   }
   
   list<PhotData*> lp;
   for( int i=1; i<=xc.length(1); ++i ) 
   {
      lp.push_back( new PhotData( xc(i), yc(i), a(i), b(i), pa(i), ir1(i),
                                  g.fixed_aper_/2. ) );
   }
   return lp;   
}


// convert iraf style linear transf. coefficients to a matrix
// 
void iraf2matrix()
{
   cerr << "  Transformation given in IRAF format." << endl;

   const float rad_per_deg = 0.017453292519;
   const float xmag = g.aptransf_[0];
   const float ymag = g.aptransf_[1];
   const float xrot = g.aptransf_[2]*rad_per_deg; 
   const float yrot = g.aptransf_[3]*rad_per_deg;

   g.aptransf_[0] =  xmag*cos(xrot);
   g.aptransf_[1] =  ymag*sin(yrot);
   g.aptransf_[2] = -xmag*sin(xrot);
   g.aptransf_[3] =  ymag*cos(yrot);
}


bool need_to_transform()
{
   return g.aptransf_[0] || g.aptransf_[1] || g.aptransf_[2] || 
      g.aptransf_[3];
}

/*
inline float hypot( const float x, const float y )
{
   return sqrt( x*x + y*y );
} 

DECLARE_BINARY_FUNC( hypot, float );
*/

void linear_transform_ap( MVector& a, MVector& b, MVector& pa )
{
   cerr << "  Transforming apertures ..." << endl;

   const float D0 = g.aptransf_[0];
   const float D1 = g.aptransf_[1];
   const float D2 = g.aptransf_[2];
   const float D3 = g.aptransf_[3];

   MVector pa_rad = pa*M_PI/180.0;

   // x and y components of new principal axes
   MVector a1x( cos(pa_rad)*D0 + sin(pa_rad)*D1 );
   MVector a1y( cos(pa_rad)*D2 + sin(pa_rad)*D3 );
   MVector b1x( sin(pa_rad)*D0 - cos(pa_rad)*D1 );
   MVector b1y( sin(pa_rad)*D2 - cos(pa_rad)*D3 );
  
   // reduce vectors to length and position angle ...
   a *= hypot( a1x, a1y );
   b *= hypot( b1x, b1y );
   pa = atan2( a1y, a1x ) * 180./M_PI; // in degrees again
}

