/*
 * psf.cpp
 * Creatte a psf template from a number of stellar images
 *
 * --------------------------------------------------------------------
 * $Id: psf.cpp 2 2002-08-19 17:00:29Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include "globals.h"
#include "yoda.h"
#include "object.h"

#include <list>
#include <cmath>

#include <ltl/marray.h>
#include <ltl/marray_io.h>
#include <ltl/statistics.h>

extern Globals g;


float image_bilinear( const MImage& image, const float x, const float y );



MImage psf( list<Object*> stars, const MImage& image )
{
   int psize = g.psftemplsize_;
   if( !(psize % 2) )
      ++psize;
   int ps2 = psize/2;

   int nstars = stars.size();

   MArray<float,3> t( psize, psize, nstars );
   t = 0.0f;
   const Range all = Range::all();

   // combine stellar images to psf template
   int n = 1;
   for( list<Object*>::iterator i=stars.begin(); i!=stars.end(); ++i ) 
   {
      Object& obj = **i;
      int xc = int(obj.icx()+.5);
      int yc = int(obj.icy()+.5);

      const Range& rx = Range(xc-ps2,xc+ps2);
      const Range& ry = Range(yc-ps2,yc+ps2);

      const Range& cx = Range(xc-2,xc+2);
      const Range& cy = Range(yc-2,yc+2);

      t(all,all,n++) = image(rx,ry) / sum( image(cx,cy) );
   }

   MImage psf( psize, psize );
   MImage psn( psize, psize );

   // kappa-sigma-average all stellar images
   for( int j=1; j<=psize; ++j )
      for( int i=1; i<=psize; ++i ) 
      {
	 double avg, d;
	 kappa_sigma_average( t( i, j, all ), 2.5, &avg, &d );
	 psf(i,j) = avg;
	 psn(i,j) = avg/d;
      }

   //cout << ma_sum( psf ) << endl;
  
   cerr.setf( ios::fixed, ios::floatfield );
   cerr.precision(2);
   cerr << "PSF :" << MArray<float,2>( psf/psf(ps2+1,ps2+1) ) << endl;
   cerr << psn << endl;

   cerr.setf( ios::floatfield, ios::fixed );
   cerr.precision(6);
   
   psf /= psf(ps2+1,ps2+1);
   return psf;
}


float image_bilinear( const MImage& image, const float x, const float y )
{
   int x1 = int(x), x2 = int(x+1.0f);
   int y1 = int(y), y2 = int(y+1.0f);

   float t = (x-x1);
   float u = (y-y1);

   return (1.0f-t)*(1.0f-u)*image(x1,y1) + 
      t*(1.0f-u)*image(x2,y1) + 
      t*u*image(x2,y2) + 
      (1.0f-t)*image(x1,y2);
}


