/* 
 * resolution.cpp
 * routine for resolution classifier
 *
 * --------------------------------------------------------------------
 * $Id: resolution.cpp 24 2003-10-29 21:04:29Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <list>
#include <iostream>
#include <iomanip>
using std::setw;

#include <ltl/marray.h>
#include <ltl/marray_io.h>
#include <ltl/statistics.h>


#include "yoda.h"
#include "object.h"
#include "photdata.h"
#include "likelihood.h"


template<class func>
bool amoeba( MArray<float,2>& p, MArray<float,1>& y, func& f, 
             const float ftol=1e-6, const int nmax=1000 );
template<class func>
float amotry( MArray<float,2>& p, MArray<float,1>& y, 
              FixedVector<float,2>& psum,
              func& f, const int ihi, const float fac );

void resolution_amoeba( const MImage& I, 
                        const float sky, const float flux, const MImage& psf,
                        float* alpha, float* beta, float* lum );



void resolution_classifier( MImage& Image, MImage& Psf, 
                            list<PhotData*>& pl, const float gain )
{
   const int wx = Psf.length(1)/2;
   const int wy = Psf.length(2)/2;

   Psf.setBase( -wx, -wy );

   list<PhotData*>::iterator i = pl.begin();
   int count=0;
   
   for( ; i!=pl.end(); ++i ) 
   {
      if( !(++count % 100) ) 
         cerr << setw(7) << count << " "; 

      PhotData& pd = **i;
      
      const int xc = int(pd.xc_ + 0.5);
      const int yc = int(pd.yc_ + 0.5);
      if( pd.f_ < 0.0 ||
          xc-wx<Image.minIndex(1) || xc+wx>Image.maxIndex(1) ||
          yc-wy<Image.minIndex(2) || yc+wy>Image.maxIndex(2) )
      {
         pd.alpha_ = pd.beta_ = pd.lum_ = 0.0;
         continue;
      }
         
      // Make sure the image and sky are positive ...
      MImage S = Image( Range(xc-wx, xc+wx), Range(yc-wy, yc+wy) );
      MImage I( S.shape() );
      
      float mmin = min(S);
      if( mmin > 0 )
         mmin = 0;
      else
         mmin = fabs(mmin) + 1.;
      
      I = (S+mmin)*gain;
      I.setBase( -wx, -wy );

      float alpha, beta, lumfac;
      
      resolution_amoeba( I, (pd.sky_+mmin)*gain, pd.f_*gain, Psf, 
                         &alpha, &beta, &lumfac );
      
      pd.alpha_ = alpha;
      pd.beta_ = beta;
      pd.lum_ = lumfac;
   }
   cerr << endl;
}


void resolution_amoeba( const MImage& I, 
                        const float sky, const float flux,
                        const MImage& psf,
                        float* alpha, float* beta, float* lum )
{
   LiklihoodFunc_Alpha A( I, sky, flux, psf );
   LiklihoodFunc_Beta  B( I, sky, flux, psf );
   
   // alpha and luminosity
   MVector y( 3 );
   y(1) = A( 1.0, 1.0 );
   y(2) = A( 1.2, 0.9 );
   y(3) = A( 1.4, 1.2 );

   MImage p( 3, 2 );
   p(1,1) = 1.0; p(1,2) = 1.0;
   p(2,1) = 1.2; p(2,2) = 0.9;
   p(3,1) = 1.4; p(3,2) = 1.2;   

   bool conv1 = amoeba( p, y, A, 1e-4 );

   float a1 = p(1,1), l1 = p(1,2), y1 = y(1);

   // beta and luminosity
   y(1) = B( 0.2, 1.0 );
   y(2) = B( 0.3, 0.9 );
   y(3) = B( 0.4, 1.2 );

   p(1,1) = 0.2; p(1,2) = 1.0;
   p(2,1) = 0.3; p(2,2) = 0.9;
   p(3,1) = 0.4; p(3,2) = 1.2;   
   bool conv2 = amoeba( p, y, B, 1e-4 );

   float y2 = y(1);

   // return [alpha, beta, luminosity]
   if( conv1 || conv2 )
   {
      if( y1 <= y2 )
      {
         *alpha = a1;
         *beta  = 0.0;
         *lum   = l1;
      }
      else
      {
         *alpha = 1.0;
         *beta  = p(1,1);
         *lum   = p(1,2);
      }
   }
   else
   {
      *alpha = 0.0;
      *beta  = 0.0;
      *lum   = 0.0;
   }
   
}


/*
     Multidim. minimization of the function f( x ), where x(1..N) is
     the vector of of parameters, by the downhill simplex method.

     On entry, y(i) has to be set to the value of function( Xi ), where Xi are
     N+1 input vectors. p is a (N+1)-by-N matrix whose columns are the Xi.
     On exit, p contains N+1 vectors all within ftol of a minimum.
*/
template<class func>
bool amoeba( MArray<float,2>& p, MArray<float,1>& y, func& f, 
             const float ftol, const int nmax )
{
   int i, ihi, ilo, inhi;
   float rtol, ysave, ytry;
   
   FixedVector<float,2> psum;
   psum(1) = p(1,1) + p(2,1) +p(3,1);
   psum(2) = p(1,2) + p(2,2) +p(3,2);
      
   int ntry = 0;

   for (;;) {
      if( ++ntry > nmax )
         return false;

      ilo=1;
      ihi = y(1)>y(2) ? (inhi=2,1) : (inhi=1,2);
      for (i=1;i<=3;i++) {
         if (y(i) <= y(ilo)) ilo=i;
         if (y(i) > y(ihi)) {
            inhi=ihi;
            ihi=i;
         } else if (y(i) > y(inhi) && i != ihi) inhi=i;
      }
      rtol=2.0*fabs(y(ihi)-y(ilo))/(fabs(y(ihi))+fabs(y(ilo)));
      if (rtol < ftol) {
         swap(y(1),y(ilo));
         swap(p(1,1),p(ilo,1));
         swap(p(1,2),p(ilo,2));
         break;
      }
      ytry=amotry(p,y,psum,f,ihi,-1.0);
      if (ytry <= y(ilo))
         ytry=amotry(p,y,psum,f,ihi,2.0);
      else if (ytry >= y(inhi)) {
         ysave=y(ihi);
         ytry=amotry(p,y,psum,f,ihi,0.5);
         if (ytry >= ysave) {
            for (i=1;i<=3;i++) {
               if (i != ilo) {
                  p(i,1)=psum(1)=0.5*(p(i,1)+p(ilo,1));
                  p(i,2)=psum(2)=0.5*(p(i,2)+p(ilo,2));
                  y(i)=f(psum(1), psum(2));
               }
            }
            psum(1) = p(1,1) + p(2,1) +p(3,1);
            psum(2) = p(1,2) + p(2,2) +p(3,2);
         }
      } 
   }
   return true;
}


template<class func>
float amotry( MArray<float,2>& p, MArray<float,1>& y, 
              FixedVector<float,2>& psum,
              func& f, const int ihi, const float fac )
{
   float fac1,fac2,ytry;

   FixedVector<float,2> ptry;

   fac1=(1.0-fac)/2;
   fac2=fac1-fac;

   ptry(1)=psum(1)*fac1-p(ihi,1)*fac2;
   ptry(2)=psum(2)*fac1-p(ihi,2)*fac2;

   ytry=f( ptry(1), ptry(2) );
   if (ytry < y(ihi)) 
   {
      y(ihi)=ytry;
      psum(1) += ptry(1)-p(ihi,1);
      psum(2) += ptry(2)-p(ihi,2);
      p(ihi,1)=ptry(1);
      p(ihi,2)=ptry(2);
   }
   return ytry;
}


/* defined in stl algobase.h ...

template<class T>
void swap( T& a, T& b )
{
   T tmp;
   tmp=a;
   a=b;
   b=tmp;
}

*/
