/*
 * split.cpp
 * Object splitting
 *
 * --------------------------------------------------------------------
 * $Id: split.cpp 35 2004-10-06 18:43:18Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include <ltl/marray.h>

#include "config.h"
#include "yoda.h"

#include "globals.h"
#include "object.h"

#include <cmath>

extern Globals g;

//typedef MArray<float,2> MImage;
//typedef MArray<char,2> MMask;

list<Object*> split( list<Object*> l )
{
   list<Object*> l2;
   int numsplit=0;

#ifdef YODA_DEBUG_SPLIT
   int c=0;
#endif

   // lower npixmin
   g.npixmin_ = g.splitnpix_;
 
   cerr << "  Minimum number of pixels above isophote for splitting: " 
	<< g.splitnpix_ << endl;
   cerr << "  Using isophote stepsize of "<<g.fracsplit_<<"*maximum flux,\n" 
	<< "  minimum isophote at "<<g.splitrange_[0] << ", maximum at "  << g.splitrange_[1] << " *maximum flux.\n";
   // for all detected objects 
   for( list<Object*>::iterator i=l.begin(); i!=l.end(); ++i ) 
   {
      float splitisop = 0.0f;
      bool split = false;
      Object& obj = **i;

      // make subimage containing object;
      MImage subimage( Range(obj.xmin(),obj.xmax()), 
		       Range(obj.ymin(),obj.ymax()) );
      MMask  mask( Range(obj.xmin(),obj.xmax()), 
		   Range(obj.ymin(),obj.ymax()) );

      float mmin = obj.maxv()*g.splitrange_[0];
      mmin = (mmin < obj.thresh() ? obj.thresh() : mmin);
      float mmax = obj.maxv()*g.splitrange_[1];

      mask = 0;
      subimage = mmin-1;
      obj.drawpixels( subimage );
    
      // ok, ... go for it, loop over increasing isophotes
      float isoph = mmin;
      float splitstep = (mmax-mmin)*g.fracsplit_; 
      if( splitstep < YODA_SPLIT_EPS )
         splitstep = YODA_SPLIT_EPS;

#ifdef YODA_DEBUG_SPLIT
      cerr << "splitting object no. " << ++c << " at " << obj.maxx() << "," << obj.maxy() << " DI:" << obj.thresh() << " step:" << splitstep << endl;
#endif

      while( isoph < mmax-splitstep ) 
      {
	 // detect at current isophote t
#ifdef YODA_DEBUG_SPLIT
         cerr << "mmax " << mmax << ", mmin " << mmin << ", step " << splitstep << endl;
         cerr << "splitting at " << isoph << "  until " << mmax << endl;
#endif
	 mask = subimage > isoph;
	 list<Object*> ll = assemble( mask, subimage, subimage, 0, true );

	 if( ll.empty() )
	    break;

	 if( ll.size() == 1 ) 
	 {
	    // still only one object ...
	    delete_list( ll );
	 } 
	 else 
	 {
	    // found second object ...
#ifdef YODA_DEBUG_SPLIT
	    cerr << "subobject found!" << endl;
#endif
	    obj.setsplit();
	    for( list<Object*>::iterator j=ll.begin(); j!=ll.end(); ++j ) 
	    {
	       // requeue new objects at end of list ...
	       Object& objnew = **j;
	       if( objnew != obj ) 
	       {
		  splitisop = isoph; //last splitting isophote
		  split = true;
		  numsplit++;
		  objnew.setthresh( isoph );
		  objnew.setpixels( subimage, mmin-1 );
		  objnew.setsplit();
		  // requeue after current object
		  i = l.insert( ++i, &objnew );
		  --i;
	       }
	    }
	 }
	 // increase isophote
	 isoph += splitstep;
      } // while ( isoph <  ...

      if( split ) 
      {
	 // now redetect the main object (in image with all other pixels zeroed)
	 // at last splitting isophote and reque at end of list
	 mask = subimage > splitisop;
	 list<Object*> ll = assemble( mask, subimage, subimage, 0, true );
         int nn = ll.size();
	 if( nn != 1 ) 
	 {
            if( nn != 0 )
            {
               cerr << "Oops, ... splitting internal error!\n";
               cerr << ll.size() << " objects found, 1 expected, " << split << endl;
               delete_list( ll );
            }
	    continue;
	 } 
	 // requeue after current object
	 Object *requeue = ll.front();
	 requeue->setthresh( splitisop );
	 requeue->setsplit();
	 i = l.insert( ++i, requeue );
	 --i;
      }
      else 
      {
	 // ok, if no splitting was neccessary anymore, copy the object to
	 // final list
	 l2.push_back( new Object( obj ) );
      }
    
   } // for( list<Object*>::iterator ...

   int n = l2.size();
   cerr << "  " << numsplit << " multiple objects splitted." << endl
	<< "  " << n << " objects detected in total." << endl;
   return l2;
}


void delete_list( list<Object*> l )
{
   for( list<Object*>::iterator i=l.begin(); i!=l.end(); ++i ) 
      delete *i;
}


